/* $Workfile: T8pBurstEx.c $       $Revision: 72 $      $Modtime: 3/12/08 12:06p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8pBurstEx.c

    A simple example console application demonstrating the use of the
    DMA Direct Sending and Receiving with the OTX Driver using the
    board's Burst device to access the data using DMA.

      Thor-2-PCI-Plus
      Thor-2-PCI-Express
      Thor-4-PCI-Plus
      Thor-8-PCI-Plus
      Sleipnir-1-PCI-Plus
      Hugin-2-PC104
      Thor-2-ExpressCard
      Thor-4-ExpressCard

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2(4,8) Lines respectively for Thor-2(4,8)-PCI-Plus which can 
    be looped with an external loop cable.  Sleipnir-1-PCI-Plus can be looped 
    back to itself.

    The direct sends and receives operate on a highway basis and are
    hosted by the Burst device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#include "OtxBurst.h"
#include "OtxBurstD.h"

#define DEMO_DISPLAY_BYTES 1   // Comments this macro out to display only the received F-bits (T1 mode only) in verbose mode

#define DEMO_DEFAULT_LOG_FILE   "T8pBurstEx.log"
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  int                    coLi;
  int                    coHighways;
  int                    coLiPerBurstDevs;
  int                    coTimeslotsOnBurstHwy;
  int                    nCurrentLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  int                    boardNo;
  OTX_BOOL               bDisableBurstTx;
  OTX_BOOL               bDisableBurstRx;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_UINT32        g_nRcvStat_Starts = 0;            // Number of actual dma transfers initiated
OTX_UINT32        g_nRcvStat_Completes = 0;         // Number of actual dma transfers completed IRQs
OTX_UINT32        g_nRcvStat_Partial = 0;           // Number of partial transfer due to framemask
OTX_UINT32        g_nRcvStat_Full = 0;              // Number of count of max dma size transfers
OTX_UINT32        g_nRcvStat_Aborts = 0;            // Number of actual dma transfers aborts
OTX_UINT32        g_nRcvStat_Amount = 0;            // Number of amount of data transfered
OTX_UINT32        g_nRcvStat_Errors = 0;            // Number of count of processing errors
OTX_UINT32        g_nRcvStat_OUruns = 0;            // Number of amount of over/under runs
OTX_UINT32        g_nRcvStat_OUrunsUser = 0;        // Number of amount of over/under runs caused by user mode
OTX_UINT32        g_nRcvStat_DmaBusy = 0;           // Number of count of DMA busy occurances
OTX_UINT32        g_nRcvStat_Notify = 0;            // Number of logical device notifies
OTX_UINT32        g_nRcvStat_NotifyPartial = 0;     // Number of logical device notifies - partial
OTX_UINT32        g_nRcvStat_FifoSize = 0;          // size of the Burst FIFO
OTX_UINT32        g_nRcvStat_FifoTrigger = 0;       // the current FIFO Trigger value
OTX_UINT32        g_nRcvStat_FifoCount = 0;         // the current amount in the FIFO
OTX_UINT32        g_nRcvStat_FifoPeak = 0;          // the peak FIFO usage
OTX_UINT32        g_nRcvStat_FifoTriggerCount = 0;  // the current FIFO trigger fire value

OTX_UINT32        g_nSendStat_Starts = 0;            // Number of actual dma transfers initiated
OTX_UINT32        g_nSendStat_Completes = 0;         // Number of actual dma transfers completed IRQs
OTX_UINT32        g_nSendStat_Partial = 0;           // Number of partial transfer due to framemask
OTX_UINT32        g_nSendStat_Full = 0;              // Number of count of max dma size transfers
OTX_UINT32        g_nSendStat_Aborts = 0;            // Number of actual dma transfers aborts
OTX_UINT32        g_nSendStat_Amount = 0;            // Number of amount of data transfered
OTX_UINT32        g_nSendStat_Errors = 0;            // Number of count of processing errors
OTX_UINT32        g_nSendStat_OUruns = 0;            // Number of amount of over/under runs
OTX_UINT32        g_nSendStat_OUrunsUser = 0;        // Number of amount of over/under runs caused by user mode
OTX_UINT32        g_nSendStat_DmaBusy = 0;           // Number of count of DMA busy occurances
OTX_UINT32        g_nSendStat_Notify = 0;            // Number of logical device notifies
OTX_UINT32        g_nSendStat_NotifyPartial = 0;     // Number of logical device notifies - partial
OTX_UINT32        g_nSendStat_FifoSize = 0;          // size of the Burst FIFO
OTX_UINT32        g_nSendStat_FifoTrigger = 0;       // the current FIFO Trigger value
OTX_UINT32        g_nSendStat_FifoCount = 0;         // the current amount in the FIFO
OTX_UINT32        g_nSendStat_FifoPeak = 0;          // the peak FIFO usage
OTX_UINT32        g_nSendStat_FifoTriggerCount = 0;  // the current FIFO trigger fire value


OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE        *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bStatsThreadActive = OTX_FALSE;
OTX_BOOL    g_bReadBurstStatsThreadActive = OTX_FALSE;


struct {
  OTX_HANDLE hThorController;
  OTX_HANDLE hLi[8];
  OTX_HANDLE hQuadLi[2];
  OTX_HANDLE hBurst;
  OTX_HANDLE hTss;
} g_hPhysDevs;

struct {
  OTX_HANDLE hLiIn[8];
  OTX_HANDLE hLiOut[8];
} g_hPipes;

// direct Burst DMA send
struct sBurstSender {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile;
  OTX_INT32   nTsToSend;  // Default is -1 (All TS)
  OTX_UINT32  cbTxBuf;
  OTX_BOOL    bFileLoop;
};

// direct Burst DMA receive
struct sBurstReceiver {
  OTX_UINT32  nSourceId;
  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE        *pRxFile;
  OTX_INT32   nTsToRecord;  // Default is -1 (All TS)
};

struct {
  struct sBurstSender mBurstSender[2];        // one for each possible highway hosted by the Burst device
  struct sBurstReceiver mBurstReceiver[2];    // one for each possible highway hosted by the Burst device
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-8 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.

    The Burst device hosts the DMA functionality.  If we open the Burst
    device, when we enable that device, the process of transferring data from
    the board to the driver will take place via DMA.  The Burst device provides
    the necessary buffering to process the data from the board until it can be
    processed and filtered to the logical devices that are hosted by Burst.
    (see DemoCreateLogicalDevices for example of connection).

  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;
  OTX_BOOL  bOpenTss = OTX_FALSE; // Not all boards have TSS device

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_SLEIPNIR_1_PCI:
      DemoLogf("Opening Sleipnir board\n");
      break;
    case OTX_DEVICE_HUGIN_2_PC104:
      DemoLogf("Opening Hugin-2-PC104 board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_PLUS:
      bOpenTss = OTX_TRUE;
      DemoLogf("Opening Thor-2-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_PCI_EXPRESS:
      DemoLogf("Opening Thor-2-PCI-Express board\n");
      break;
    case OTX_DEVICE_THOR_4_PCI_PLUS:
      bOpenTss = OTX_TRUE;
      DemoLogf("Opening Thor-4-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_8_PCI_PLUS:
      bOpenTss = OTX_TRUE;
      DemoLogf("Opening Thor-8-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_1_PCI_PLUS:
      bOpenTss = OTX_TRUE;
      DemoLogf("Opening Thor-1-PCI-Plus board\n");
      break;
    case OTX_DEVICE_THOR_2_EXPRESSCARD:
      DemoLogf("Opening Thor-2-ExpressCard board\n");
      break;
    case OTX_DEVICE_THOR_4_EXPRESSCARD:
      DemoLogf("Opening Thor-4-ExpressCard board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }

  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_EXPRESS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the 2 individual T1/E1 line interface FALC 
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Get handles to the one quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    for (i = 0; i < 2; i++) {
      // Get handles to the two quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
    for (i = 0; i < 1; i++) {
      // Get handles to the dial T1/E1 line interface device
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DUAL_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 2 individual T1/E1 line interface members within the Dual device
      for (j = 0; j < 2; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j, g_hEventQueue, &g_hPhysDevs.hLi[j]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

    // we need to tell the driver to source the Falc from the Burst device (and not from the DSP, which is the default).
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    for (i = 0; i < 1; i++) {
      // Get handles to the one quad T1/E1 line interface devices
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_QUAD_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hQuadLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Thor-4-ExpressCard by default is set to receive on the four lines 
      // (monitor mode) with all transmitters disabled (tristate). This 
      // demo however needs the card to support the transmit data so the 
      // card will use 2 receivers and the 2 transmitters.
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hThorController, 
                                           OTX_ATTR_BRD_MONITOR_ONLY, 
                                           OTX_FALSE); 
      DemoCheckRc(g_hPhysDevs.hThorController, result);
      DemoLogf("Configure Thor-4-ExpressCard for data transmit and receive\n");

      // we'll source the Falc from the Burst device, and source the burst device from the FALC
      //  Falc <-- Burst Send (P2S)
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);
      //  Burst Recv (S2P) <-- Falc
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC);  DemoCheckRc(g_hPhysDevs.hThorController, result);

      // Get handles to the 4 individual T1/E1 line interface members within the Quad device
      for (j = 0; j < 4; j++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadLi[i], OTX_DEVICE_LI_T1E1, j, j + i*4, g_hEventQueue, &g_hPhysDevs.hLi[j + i*4]); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
      }
    }

  } else if (g_demoEnv.nBoardType == OTX_DEVICE_SLEIPNIR_1_PCI || g_demoEnv.nBoardType == OTX_DEVICE_HUGIN_2_PC104) {

    // Open the Line interfaces
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

    // we'll source the Falc from the Burst device, and source the burst device from the FALC
    //  Falc <-- Burst Send (P2S)
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(g_hPhysDevs.hThorController, result);
    //  Burst Recv (S2P) <-- Falc
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_S2P_SOURCED, OTX_BRD_S2P_SOURCED_FALC); DemoCheckRc(g_hPhysDevs.hThorController, result);
  
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_1_PCI_PLUS) {
    for (i = 0; i < 1; i++) {
      // Get handles to individual T1/E1 line interface FALC 
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController, result);
    }

  }

  // Get a handle to the physical Burst device.
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(g_hPhysDevs.hThorController, result);

  // set device buffer event trigger levels
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(g_hPhysDevs.hBurst, result);

  if(bOpenTss) {
    // Get a handle to the Time-Space Switch (TSS) on the board
    DemoLogf("Open OTX_DEVICE_TSS\n");
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_TSS, 0, 0, OTX_INVALID_HANDLE_VALUE, &g_hPhysDevs.hTss); DemoCheckRc(g_hPhysDevs.hTss, result);
  }
  else {
    g_hPhysDevs.hTss = OTX_INVALID_HANDLE_VALUE;
  }

  // Optionally (mainly for testing, or for improving system performance for single direction burst applications,
  // disable one direction of the burst device
  if (g_demoEnv.bDisableBurstRx) {
    DemoLogf("Burst RX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECV_DISABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  }
  if (g_demoEnv.bDisableBurstTx) {
    DemoLogf("Burst TX interrupts will be turned off when the Burst device is enabled\n");
    result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SEND_DISABLED,  OTX_TRUE); DemoCheckRc(g_hPhysDevs.hBurst, result);
  }
}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
static void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData(&(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // This configuration did not work for HUGIN in loop back.
  // Modified code below was taken from the ThorDemo. 
  if(g_demoEnv.nBoardType != OTX_DEVICE_HUGIN_2_PC104) {
  
    for (i = 0; i < g_demoEnv.coLi; i++) {
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) { 
        // Configure T1 data
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      } else {
        // Configure E1 data
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      
      }
    
      // Set the T1 or E1 mode
      result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    
      // Set line termination mode
      {
      
        DemoLogf("Li[%d]: Setting Line Termination mode to ", i);
        switch (g_demoEnv.eTermMode) {
        case OTX_T1E1_TERM_120_OHM:
          DemoLogf("OTX_T1E1_TERM_120_OHM\n");
          break;
        case OTX_T1E1_TERM_100_OHM:
          DemoLogf("OTX_T1E1_TERM_100_OHM\n");
          break;
        case OTX_T1E1_TERM_75_OHM:
          DemoLogf("OTX_T1E1_TERM_75_OHM\n");
          break;
        case OTX_T1E1_TERM_HIGHZ:
          DemoLogf("OTX_T1E1_TERM_HIGHZ\n");
          break;
        default:
          DemoLogf("Invalid Termination mode.\n");
          break;
        }
      
        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

        if (g_demoEnv.eTermMode == OTX_T1E1_TERM_HIGHZ) {
          DemoLogf("Li[%d]: Tristating Transmit pins in High-Z Termination mode", i);

          // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
          result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        } else {
          if(g_demoEnv.nBoardType != OTX_DEVICE_THOR_4_EXPRESSCARD) {
            // Drive the transmit pins of the Line interface in all other termination modes
            result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
          }
          else {
            if(i < 2) {
              // Drive the transmit pins of the Line interface in all other termination modes
              result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_TRUE); DemoCheckRc(g_hPhysDevs.hLi[i], result);
            }
          }
        }
      
      }
    
    
      // Set receive threshold
      DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
      { // Read back the threshold
        OTX_UINT32 nRxThreshold;
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
      }
    
      // Set Monitor (-20dB) mode
      DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
      result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    
      // Turning front end Amplifier ON/OFF      
      DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    }
    if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
    
      OTX_INT32 x;
      OTX_RESULT otxResult;

      // Li2 and li3 are receive only
      for (x = 2; x < g_demoEnv.coLi; x++) {
        otxResult = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[x], 
                                          OTX_FALSE); 
        DemoCheckRc(g_hPhysDevs.hLi[x], otxResult);
        DemoLogf("  OtxT1E1XmitPinControl for Line #%d set FALSE (Tristate)\n", x);

        result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[x], 
                                           OTX_T1E1_TERM_HIGHZ); 
        DemoCheckRc(g_hPhysDevs.hLi[x], result);

        DemoLogf("  OTX_T1E1_TERM_HIGHZ\n");
      }
    }
  }
  else {

    // Configure Transceivers
    for (i = 0; i < g_demoEnv.coLi; i++) {

      // Set termination for the T1/E1 Front End
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
        // T1 mode
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      } else {
        // E1 mode
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
      }
      result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
  }

}


void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable g_hPhysDevs.hThorController\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController, result);

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {

    for (i = 0; i < 1; i++) {
      // Enable DualFalc
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {

    for (i = 0; i < 1; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {

    for (i = 0; i < 1; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {

    for (i = 0; i < 2; i++) {
      // Enable QuadFalcs
      DemoLogf("Enable g_hPhysDevs.hQuadLi[%d]\n", i);
      result = OtxDrvEnable(g_hPhysDevs.hQuadLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadLi[i], result);
    }

  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    // Enable Physical Devices, Li's
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{

  OTX_INT32 nLi;
  OTX_RESULT result;
  OTX_UINT32 iHighwayNo;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_UINT8 sqBurstTimeSlotMask[OTX_PIPE_TIMESLOTS_MAX];

  if ((g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) ||
      (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD)) {
    OTX_HANDLE hPipe;

    // Although the timeslots connections are hardcoded on this card, we still need to connect
    // to the timeslots of the Li devices that we want to control. If we don't connect any pipes 
    // to the Li devices then it will send an Idle code on each timeslot.

    memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));

    result = OtxDrvCreatePipe(64*32, &hPipe);
    DemoCheckRc(hPipe, result);
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[0],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(hPipe, result);

    result = OtxDrvCreatePipe(64*32, &hPipe);
    DemoCheckRc(hPipe, result);
    result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[1],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(hPipe, result);

    return; // No need to connect the other end of the pipe since the pipe 
            // connections are hard-coded on this board.
  }

  // Create the Pipes (one for each direction) between the Burst Device and the LIs
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    OTX_INT32 nTs;

    // we must define the highway to use [0-1]
    iHighwayNo = nLi / g_demoEnv.coLiPerBurstDevs;

    // Create pipes (one for each direction) Use all timeslots.
    memset(sqTimeSlotMask, 0, sizeof(sqTimeSlotMask) * sizeof(sqTimeSlotMask[0]));
    memset(sqBurstTimeSlotMask, 0, sizeof(sqBurstTimeSlotMask) * sizeof(sqBurstTimeSlotMask[0]));

    DemoLogf("Creating pipe for Burst Sender/Receiver\n");

    for (nTs = 0; nTs < 32; nTs++) {
      
      sqTimeSlotMask[nTs] = 0xFF;
      
      // for this demo, we'll lay out the burst time slot mask as follows:
      // OTX_DEVICE_SLEIPNIR_1_PCI:
      //    [ 0..31]  - LI0:TS0-TS31
      //
      // OTX_DEVICE_HUGIN_2_PC104:
      // OTX_DEVICE_THOR_2_PCI_PLUS:
      // OTX_DEVICE_THOR_2_PCI_EXPRESS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0) and LI1:TS0-TS31 (hwy1) 
      // 
      // OTX_DEVICE_THOR_8_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0)  and LI4:TS0-TS31 (hwy1) 
      //    [32..63]  - LI1:TS0-TS31 (hwy0)  and LI5:TS0-TS31 (hwy1) 
      //    [64..95]  - LI2:TS0-TS31 (hwy0)  and LI6:TS0-TS31 (hwy1) 
      //    [96..127] - LI3:TS0-TS31 (hwy0)  and LI7:TS0-TS31 (hwy1) 
      //
      // OTX_DEVICE_GIMLE_16_PCI_PLUS:
      //    [ 0..31]  - LI0:TS0-TS31 (hwy0)  and LI4:TS0-TS31 (hwy1)  and LI8:TS0-TS31  (hwy2)  and LI8:TS0-TS31  (hwy3)
      //    [32..63]  - LI1:TS0-TS31 (hwy0)  and LI5:TS0-TS31 (hwy1)  and LI9:TS0-TS31  (hwy2)  and LI9:TS0-TS31  (hwy3)
      //    [64..95]  - LI2:TS0-TS31 (hwy0)  and LI6:TS0-TS31 (hwy1)  and LI10:TS0-TS31 (hwy2)  and LI10:TS0-TS31 (hwy3)
      //    [96..127] - LI3:TS0-TS31 (hwy0)  and LI7:TS0-TS31 (hwy1)  and LI11:TS0-TS31 (hwy2)  and LI11:TS0-TS31 (hwy3)
      //
      // For OTX_DEVICE_THOR_8_PCI_PLUS and OTX_DEVICE_GIMLE_16_PCI_PLUS 
      // upon connecting the pipe (see below), we have to pass the highway
      // number (0,1 for Thor-8 and 0,1,2,3 for Gimle-16). There is one highway 
      // per each QuadFalc (128TS per highway). We also pass the size of the highway,
      // which is 128 (TS) for these two boards.
      //
      // For OTX_DEVICE_SLEIPNIR_1_PCI, OTX_DEVICE_HUGIN_2_PC104,  
      // OTX_DEVICE_THOR_2_PCI_PLUS, and OTX_DEVICE_THOR_2_PCI_EXPRESS,
      // the size of the highway is 32.
      //
      // this is just a simple algorithm to accomplish the above
      sqBurstTimeSlotMask[nTs + ((nLi % g_demoEnv.coLiPerBurstDevs) * 32)] = 0xFF;
      
    }


    // --------------------------------------------------------------------
    // Create an incoming and outgoing Pipe representing the external T1/E1
    // --------------------------------------------------------------------
    result = OtxDrvCreatePipe(64*32, &g_hPipes.hLiIn[nLi]);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);
    result = OtxDrvCreatePipe(64*32, &g_hPipes.hLiOut[nLi]);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    // -------------------------
    // Connect pipe: Burst -> Li
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_INPUT,  g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.coTimeslotsOnBurstHwy);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiIn[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiIn[nLi], result);

    // -------------------------
    // Connect pipe: Li -> Burst
    // -------------------------
    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

    result = OtxDrvConnectPipe(g_hPipes.hLiOut[nLi], OTX_PIPE_OUTPUT, g_hPhysDevs.hBurst, iHighwayNo, sqBurstTimeSlotMask, g_demoEnv.coTimeslotsOnBurstHwy);
    DemoCheckRc(g_hPipes.hLiOut[nLi], result);

  }

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);
}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice

      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType);
      }

      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly);
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber);
      }

      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }

      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
#if 0
      {
        // API call to get the result code data
        OTX_UINT32 Result, Severity, SourceId, LineNum;

        OtxResultCode(result, &Result, &Severity, &SourceId, &LineNum);
        DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", Result, SourceId, LineNum, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      }
#endif

    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              continue;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              DemoBurstSender(eventData.m_nRequestId);
              continue;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.
*******************************************************************************/
void *DemoStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT        result;
  OTX_DATETIME      nTimeStamp;
  OTX_UINT32        stat_Starts;            // Number of actual dma transfers initiated
  OTX_UINT32        stat_Completes;         // Number of actual dma transfers completed IRQ's
  OTX_UINT32        stat_Partial;           // Number of partial transfer due to framemask
  OTX_UINT32        stat_Full;              // Number of count of max dma size transfers
  OTX_UINT32        stat_Aborts;            // Number of actual dma transfers aborts
  OTX_UINT32        stat_Amount;            // Number of amount of data transfered
  OTX_UINT32        stat_Errors;            // Number of count of processing errors
  OTX_UINT32        stat_OUruns;            // Number of amount of over/under runs
  OTX_UINT32        stat_OUrunsUser;        // Number of amount of over/under runs caused by user mode
  OTX_UINT32        stat_DmaBusy;           // Number of count of DMA busy occurances
  OTX_UINT32        stat_Notify;            // Number of logical device notifies
  OTX_UINT32        stat_NotifyPartial;     // Number of logical device notifies - partial
  OTX_UINT32        stat_FifoSize;          // size of the Burst FIFO
  OTX_UINT32        stat_FifoTrigger;       // the current FIFO Trigger value
  OTX_UINT32        stat_FifoCount;         // the current amount in the FIFO
  OTX_UINT32        stat_FifoPeak;          // the peak FIFO usage
  OTX_UINT32        stat_FifoTriggerCount;  // the current FIFO trigger fire value

#define STATS_SHOW_FRAME_REF
#ifdef STATS_SHOW_FRAME_REF
  OTX_UINT32                ui32;
  OtxBurstFrameRefAndTime   FrameRef;             // frame reference data
  static OTX_UINT64         nLastFrame = 0;
#endif  // #ifdef STATS_SHOW_FRAME_REF
  

  while (g_bStatsThreadActive) {
    OtxDrvGetCurrentTime(&nTimeStamp);
    
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS,             &stat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES,          &stat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL,            &stat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL,               &stat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS,             &stat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT,             &stat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS,             &stat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS,             &stat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER,        &stat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY,            &stat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY,             &stat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL,     &stat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE,           &stat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,        &stat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_COUNT,          &stat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_PEAK,           &stat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER_COUNT,  &stat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Recv:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d, FIFO Size:%d, Trigger:%d, Count:%d, Peak:%d, TriggerCount:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial,
        stat_FifoSize,
        stat_FifoTrigger,
        stat_FifoCount,
        stat_FifoPeak,
        stat_FifoTriggerCount
    );

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS,             &stat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES,          &stat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL,            &stat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL,               &stat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS,             &stat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT,             &stat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS,             &stat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS,             &stat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER,        &stat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY,            &stat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY,             &stat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL,     &stat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_SIZE,             &stat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,          &stat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_COUNT,            &stat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_PEAK,             &stat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER_COUNT,    &stat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);

    DemoPrintTimeStamp(nTimeStamp);
    DemoLogf("Burst Send:: Starts:%d, Completes:%d, Partial:=%d, Full=%d, Aborts:%d, Amount:%x, Errors:%d, OUErrors:%d, OUErrUser:%d, DMABusy:%d, Notify:%d, NotifyPartial:%d, FIFO Size:%d, Trigger:%d, Count:%d, Peak:%d, TriggerCount:%d\n",
        stat_Starts,
        stat_Completes,
        stat_Partial,
        stat_Full,
        stat_Aborts,
        stat_Amount,
        stat_Errors,
        stat_OUruns,
        stat_OUrunsUser,
        stat_DmaBusy,
        stat_Notify,
        stat_NotifyPartial,
        stat_FifoSize,
        stat_FifoTrigger,
        stat_FifoCount,
        stat_FifoPeak,
        stat_FifoTriggerCount
    );

#ifdef STATS_SHOW_FRAME_REF
    result = OtxDrvGetAttributeValue(g_hPhysDevs.hBurst, OTX_ATTR_BURST_GET_FRAME_REF_DATA, &FrameRef, sizeof(FrameRef), &ui32); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // just skip if we wrapped
    if (FrameRef.m_FrameCounter > nLastFrame) {
      DemoPrintTimeStamp(nTimeStamp);
      DemoLogf("Burst FRef:: %d fps  Cntr:%08x%08x  Time:%08x%08x ( ",
        nLastFrame == 0 ? 0 : (OTX_UINT32) (FrameRef.m_FrameCounter - nLastFrame),
        (OTX_UINT32) (FrameRef.m_FrameCounter >> 32),
        (OTX_UINT32) (FrameRef.m_FrameCounter),
        FrameRef.m_FrameDateTime.sUint32.nHigh,
        FrameRef.m_FrameDateTime.sUint32.nLow
      );
      DemoPrintTimeStamp(FrameRef.m_FrameDateTime);
      DemoLogf(")\n");
    }
    nLastFrame = FrameRef.m_FrameCounter;
#endif  // #ifdef STATS_SHOW_FRAME_REF

    OsaThreadSleep(1000);
  }

  DemoLogf("DemoStatsThread() terminating.\n");

  return 0;
}



/*******************************************************************************
  @func
    Print DMA statistics.

    These stats are obtained from the driver, and are being used during
    initial release as a debugging feature.  These stats may be eliminated
    in the future.
*******************************************************************************/
void *DemoReadBurstStatsThread(
  void *lpParameter   // thread data
)
{
  OTX_RESULT        result;


  while (g_bReadBurstStatsThreadActive) {
    
    /*
    ** Get burst receive status.
    ** Request the first three stats are updated at a 2 to 4 mSec rate,
    ** reading the values quickly to keep them from changing too much
    ** between the first one read and the last.
    */
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_STARTS,             &g_nRcvStat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_COMPLETES,          &g_nRcvStat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_PARTIAL,            &g_nRcvStat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_FULL,               &g_nRcvStat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ABORTS,             &g_nRcvStat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_AMOUNT,             &g_nRcvStat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_ERRORS,             &g_nRcvStat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS,             &g_nRcvStat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_OURUNS_USER,        &g_nRcvStat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_DMABUSY,            &g_nRcvStat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY,             &g_nRcvStat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_RECV_NOTIFY_PARTIAL,     &g_nRcvStat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_SIZE,           &g_nRcvStat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,        &g_nRcvStat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_COUNT,          &g_nRcvStat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_PEAK,           &g_nRcvStat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER_COUNT,  &g_nRcvStat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    /*
    ** Get burst send status
    ** Request the first three stats are updated at a 2 to 4 mSec rate,
    ** reading the values quickly to keep them from changing too much
    ** between the first one read and the last.
    */
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_STARTS,             &g_nSendStat_Starts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_COMPLETES,          &g_nSendStat_Completes         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_PARTIAL,            &g_nSendStat_Partial           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_FULL,               &g_nSendStat_Full              ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ABORTS,             &g_nSendStat_Aborts            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_AMOUNT,             &g_nSendStat_Amount            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_ERRORS,             &g_nSendStat_Errors            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS,             &g_nSendStat_OUruns            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_OURUNS_USER,        &g_nSendStat_OUrunsUser        ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_DMABUSY,            &g_nSendStat_DmaBusy           ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY,             &g_nSendStat_Notify            ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_STAT_SEND_NOTIFY_PARTIAL,     &g_nSendStat_NotifyPartial     ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_SIZE,             &g_nSendStat_FifoSize          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,          &g_nSendStat_FifoTrigger       ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_COUNT,            &g_nSendStat_FifoCount         ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_PEAK,             &g_nSendStat_FifoPeak          ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER_COUNT,    &g_nSendStat_FifoTriggerCount  ); DemoCheckRc(g_hPhysDevs.hBurst, result);
    // Let other task run, getting status is not the highest priority
    OsaThreadSleep(10);

    // Read values about once per second
    OsaThreadSleep(1000);
  }

  DemoLogf("DemoReadBurstStatsThread() terminating.\n");

  return 0;
}

/*******************************************************************************
  @func
    Burst Sender Function
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_UINT32  ui32;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

START:
  if (g_hLogicDevs.mBurstSender[nSourceId].pTxFile) {
    // Prime the transmit buffer with 0xFF (only really needed for single TS transmission)
    memset(&g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0], 0xFF, sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]));

    if (g_hLogicDevs.mBurstSender[nSourceId].nTsToSend == -1) {
      // All TS
      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = fread(
        &g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf,
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[0]),
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE / 32]; // On some boards this is bigger than we actually use
      int cbSingleTsBuf;
      int byteCount = 0;
      int cbMaxSingleTsBuf;
      int x;

      cbMaxSingleTsBuf = DEMO_BURST_BUF_SIZE / g_demoEnv.coTimeslotsOnBurstHwy / sizeof(singleTsBuf[0]);

      cbSingleTsBuf = fread(
        &singleTsBuf,
        sizeof(singleTsBuf[0]),
        cbMaxSingleTsBuf,
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile
        );

      if (cbSingleTsBuf > 0) {
        if ((cbSingleTsBuf > 0) && (cbSingleTsBuf < cbMaxSingleTsBuf)) {
          // We are at the end of the file, and we don't have a full buffer to send.
          // Pad with 0xFF at the end of the buffer to make sure that we don't underrun the Burst Sender
          memset(&singleTsBuf[cbSingleTsBuf], 0xFF, (cbMaxSingleTsBuf - cbSingleTsBuf) / sizeof(singleTsBuf[0]) );
          cbSingleTsBuf = cbMaxSingleTsBuf;
        }
        
        // Dispurse the single TS to be transmitted in the whole transmit buffer 
        byteCount = 0;
        //DemoLogf("g_demoEnv.nCurrentLi=%d, g_demoEnv.coLi=%d, g_demoEnv.coHighways=%d, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend=%d \n", g_demoEnv.nCurrentLi, g_demoEnv.coLi, g_demoEnv.coHighways, g_hLogicDevs.mBurstSender[nSourceId].nTsToSend);

        if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
          // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots.
          // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots.
          // So the standard formula for calculating where to stuff the timeslots into the
          // burst highway does not apply to this board.
          // So calculate it separately.
          for ( 
            x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend*4 + g_demoEnv.nCurrentLi*2; 
          (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
          x += 128
            ) {
            g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
          }
        } 
        else
        if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
          // The Li0 timeslots on this card are the TS0, TS4, TS8,  ... , TS124 timeslots.
          // The Li1 timeslots on this card are the TS1, TS5, TS9,  ... , TS125 timeslots.
          // The Li2 is receive only send not supported. Timeslots are not used.
          // The Li3 is receive only send not supported. Timeslots are not used.
          // So the standard formula for calculating where to stuff the timeslots into the
          // burst highway does not apply to this board.
          // So calculate it separately.
          for ( 
            x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend*4 + g_demoEnv.nCurrentLi; 
            (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
            x += 128
            ) {
            g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
          }
        }
        else {
          
          for ( 
            x = g_hLogicDevs.mBurstSender[nSourceId].nTsToSend; 
          (byteCount < cbSingleTsBuf) && (x < DEMO_BURST_BUF_SIZE); 
          x += g_demoEnv.coTimeslotsOnBurstHwy
            ) {
            g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf[x] = singleTsBuf[byteCount++];
          }
        }

        
        assert(byteCount == cbSingleTsBuf);
          
      }

      g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = cbSingleTsBuf * g_demoEnv.coTimeslotsOnBurstHwy;

    }
    
    if (g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf > 0) {
      ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;

      assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

      otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
      DemoCheckRc(g_hPhysDevs.hBurst, otxRc);
      
      // check the amount buffered w/ how much we requested to be buffered
      //  (since we set the trigger amount to the buffer size, there should
      //    have been room to buffer this, so these two should be the same).
      //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      if (ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
        DemoLogf("DemoBurstSender[%d]: ui32 (%d) != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf (%d)\n", nSourceId, ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      }
      
    } else {
      
      if (g_hLogicDevs.mBurstSender[nSourceId].bFileLoop) {
        if (g_hLogicDevs.mBurstSender[nSourceId].bVerbose) {
          DemoLogf("DemoBurstSender[%d]: Transmission complete. Starting file from the beginning.\n", nSourceId);
        }
        rewind(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        goto START;
      } else {
        DemoLogf("DemoBurstSender[%d]: Transmission complete. Closing file\n", nSourceId);
        fclose(g_hLogicDevs.mBurstSender[nSourceId].pTxFile);
        g_hLogicDevs.mBurstSender[nSourceId].pTxFile = NULL;
        g_hLogicDevs.mBurstSender[nSourceId].bActive = OTX_FALSE;
      }
    }
    
  } else {
    // Send a text string (continously until aborted)

    // We MUST round amount to send to be bounded on the number of TS the highway hosts.
    //  (the buffer is representative of a burst DMA buffer for all TS on the highway hosted, TS0-TSn)
    g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf = ((g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf / g_demoEnv.coTimeslotsOnBurstHwy) * g_demoEnv.coTimeslotsOnBurstHwy);
    assert(g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf == DEMO_BURST_BUF_SIZE);

    ui32 = g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf;
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender[nSourceId].sqTxBuf, &ui32);
    DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

    // check the amount buffered w/ how much we requested to be buffered
    //  (since we set the trigger amount to the buffer size, there should
    //    have been room to buffer this, so these two should be the same).
    //assert(ui32 == g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
    if(ui32 != g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf) {
      DemoLogf("DemoBurstSender[%d]: OtxBurstSendData did not copy entire buffer.\n", nSourceId);
      DemoLogf("   Buffer size sent %d\n", g_hLogicDevs.mBurstSender[nSourceId].cbTxBuf);
      DemoLogf("   Buffer size returned %d\n", ui32);
    }
  }

  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function
*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;

  // validate parameters
  if (nSourceId < 0 || nSourceId > 2) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstReceiver[nSourceId].bActive) {
    // we're not active, so just return
    return;
  }

  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(g_hPhysDevs.hBurst, otxRc);

  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  assert(ui32 == sizeof(buf));

  // echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver[nSourceId].bVerbose) {
    int x;

#ifdef DEMO_DISPLAY_BYTES // Display the first 14 bytes of from the current Li
    DemoPrintTimeStamp(dtFrameRefTime);

    // note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("BurstReceiver[%d][%d] (len=%.5d) ID:%08x ", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf), (OTX_UINT32) (nFrameRefId & 0xffffffff));

    for (x = 0; (x < 14) && (x < (OTX_INT32) sizeof(buf)); x++) {
      // display the highway's currentLi's data
      if((g_demoEnv.nBoardType == OTX_DEVICE_HUGIN_2_PC104) ||
         (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_PLUS) || 
         (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_PCI_EXPRESS)) {
        DemoLogf("%.2X ", buf[(x + (x * 32)) + (g_demoEnv.nCurrentLi * 32)]);
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
        DemoLogf("%.2X ", buf[(g_demoEnv.nCurrentLi * 2) + x*4]);
      } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
        DemoLogf("%.2X ", buf[(g_demoEnv.nCurrentLi) + x*4]);
      } else {
        DemoLogf("%.2X ", buf[x + ((g_demoEnv.nCurrentLi % g_demoEnv.coLiPerBurstDevs) * 32)]);
      }
    }
    DemoLogf("\n");
#else 
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // Display all of the F-bits of the current Li
      int bitCount = 0;
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) ", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf));
      for (x = (g_demoEnv.nCurrentLi % g_demoEnv.coLiPerBurstDevs); x < (OTX_INT32) sizeof(buf); x += 32) {
        // Display F-bit
        DemoLogf("%.1X", buf[x] & 0x01);
        bitCount++;
           
        // Group by 4 bits to make it easier to decypher the FAS bits (001011) within the multiframe
        if ((bitCount % 4) == 0) {
          DemoLogf(" ");
        }
      }
      DemoLogf("\n");
    } else {
      // Display the first bit of TS0 (sync timeslot) of the even frames, and then the odd frames
      // In this way we can more easily decypher the frame alignment pattern (001011xx).
      // It will be repeated in either the even or the odd frames.

      // Display even frames
      int byteCount = 0;
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) (1)", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf));
      for (x = (g_demoEnv.nCurrentLi % g_demoEnv.coLiPerBurstDevs); x < (OTX_INT32) sizeof(buf); x += 32) {
        // Display Sync bit of TS0 (MSB)
        if ((byteCount++ % 2) == 0) {
          continue;
        }
        DemoLogf("%.1X", (buf[x]>>7) & 0x01);

        // Group by 4 bytes to make it easier to decypher the synchronization bits (001011xx) within the multiframe
        if ((byteCount % 8) == 0) {
          DemoLogf(" ");
        }
      }
      DemoLogf("\n");
      
      // Display odd frames
      byteCount = 0;
      DemoLogf("BurstReceiver[%d][%d] (len=%.4d) (2)", nSourceId, g_demoEnv.nCurrentLi, sizeof(buf));
      for (x = (g_demoEnv.nCurrentLi % g_demoEnv.coLiPerBurstDevs); x < (OTX_INT32) sizeof(buf); x += 32) {
        // Display Sync bit of TS0 (MSB)
        if ((byteCount++ % 2) == 1) {
          continue;
        }
        DemoLogf("%.1X", (buf[x]>>7) & 0x01);

        // Group by 4 bytes to make it easier to decypher the synchronization bits (001011xx) within the multiframe
        if (((byteCount+1) % 8) == 0) {
          DemoLogf(" ");
        }
      }
      DemoLogf("\n");

    }
#endif
  }

  // write data to a file if desired
  if (g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile) {
    int cbWritten;

    if (g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord == -1) {
      // All TS
      cbWritten = fwrite(buf, sizeof(buf[0]), sizeof(buf), g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != (int) sizeof(buf)) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", nSourceId, sizeof(buf), cbWritten);
      }
    } else {
      // Single TS
      OTX_UINT8 singleTsBuf[DEMO_BURST_BUF_SIZE/32];
      int byteCount = 0;
      int x;

      byteCount = 0;

      for (
          x = g_hLogicDevs.mBurstReceiver[nSourceId].nTsToRecord;
          x < (OTX_INT32) sizeof(buf); 
          x += g_demoEnv.coTimeslotsOnBurstHwy
      ) {
        singleTsBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleTsBuf, sizeof(singleTsBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver[nSourceId].pRxFile);
      if (cbWritten != byteCount) {
        DemoLogf("BurstReceiver[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      } else {
        //DemoLogf("BurstReceiver[%d] (byteCount=%d, cbWritten=%d) \n", nSourceId, byteCount, cbWritten);
      }
    }
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf(" a   Abort Burst[%d] Senders and Receivers\n",  g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" r   Start Burst Receiver[%d] (no file logging)\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" R   Start Burst Receiver[%d] Prompt for file to recv\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" s   Start Burst Sender[%d]. Continously send a text string\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" S   Start Burst Sender[%d]. Prompt for file to send\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" L   Burst Sender[%d]. Toggle loop file sends (currently %s)\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs, g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bFileLoop ? "On" : "Off");
    DemoLogf(" e   Enable Burst\n");
    DemoLogf(" d   Disable Burst\n");
    DemoLogf(" v   Toggle verbose (print received data) setting Burst Receiver[%d]\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" V   Toggle verbose (print send progress) setting Burst Sender[%d]\n", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    DemoLogf(" l   Query Line Status of all Lis\n");
    DemoLogf(" t   Toggle verbose buffer usage statistics\n");
    DemoLogf(" f   Forcing resynchronization\n");
    DemoLogf(" F   Reset all lines\n");
    DemoLogf(" T   Reset buffer usage statistics\n");
    DemoLogf(" N   Set burst notification value (1-32)\n");
    DemoLogf(" p   Set Loopback mode for Lis\n");
    DemoLogf(" i/I Read/Write QuadE1T1 device %d\n", g_demoEnv.nCurrentLi);
    DemoLogf(" j/J Read/Write board controller 16 bit device %d\n", g_demoEnv.nCurrentLi);
    DemoLogf(" k/K Read/Write Li device %d\n", g_demoEnv.nCurrentLi);
    DemoLogf(" m/M Read/Write board controller 32 bit device %d\n", g_demoEnv.nCurrentLi);

    break;

#if 0
  case '9':
    {
      OTX_BOOL bMonitor;
      OTX_HANDLE nBoardHandle = g_hPhysDevs.hThorController;

      result = OtxDrvGetAttributeValueBool(nBoardHandle, 
                                           OTX_ATTR_BRD_MONITOR_ONLY, 
                                           &bMonitor); 
      DemoCheckRc(nBoardHandle, result);


      bMonitor = !bMonitor;

      result = OtxDrvSetAttributeValueBool(nBoardHandle, 
                                           OTX_ATTR_BRD_MONITOR_ONLY, 
                                           bMonitor); 
      DemoCheckRc(nBoardHandle, result);

      DemoLogf("OTX_ATTR_BRD_MONITOR_ONLY for data transmit %d\n", bMonitor);

    }

    break;
#endif

  case 'e':
    // by enabling the Burst device, we start the process of transferring data
    // from the hardware to the driver via DMA.  The data is stored in the Burst
    // device's buffer for later processing to connected logical devices.
    DemoLogf("Enable g_hPhysDevs.hBurst\n");
    result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 'd':
    // by disabling the Burst device, we stop the process of transferring data
    // from the hardware to the driver via DMA.
    DemoLogf("Disable g_hPhysDevs.hBurst\n");
    result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    DemoCheckRc(g_hPhysDevs.hBurst, result);
    break;

  case 't':
    g_bStatsThreadActive = !g_bStatsThreadActive;
    if (g_bStatsThreadActive) {
      DemoLogf("Starting Stats Thread\n");
      OsaThreadStart(DemoStatsThread);
    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'F':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Reseting Li #%d... ", nLi);
        result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'v':
    g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bVerbose = !g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Receiver[%d] Thread\n", (g_hLogicDevs.mBurstReceiver[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    break;

  case 'V':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bVerbose = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bVerbose;
    DemoLogf("Setting verbose setting to %s for Burst Sender[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bVerbose)?"ON":"OFF", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    break;

  case 'a':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      }
      g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_FALSE;

      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("mBurstSender[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile);
        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;
      }
      g_hLogicDevs.mBurstSender[nHighwayNo].bActive   = OTX_FALSE;
    }
    break;

  case 'r':
    {
      OTX_UINT32 nHighwayNo;

      nHighwayNo = g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs;
      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE) {
        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else if (g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile) {
        DemoLogf("DemoBurstReceiver[%d]: Closing file\n", nHighwayNo);
        fclose(g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile);
        g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = NULL;
      } else {
        DemoLogf("Burst Receiver[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'R':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToRecord = -1;

      nHighwayNo = g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs;
      if (g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive == OTX_FALSE || g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile == NULL) {
        DemoLogf("BurstReceiver[%d]. Please enter file to receive: ", nHighwayNo);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        }

        if ((g_hLogicDevs.mBurstReceiver[nHighwayNo].pRxFile = fopen(sqzBuf, "wb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("BurstReceiver[%d]. Please enter single TS[0-%d] to receive to file (-1=All) [%d]: ", nHighwayNo, g_demoEnv.coTimeslotsOnBurstHwy - 1, nTsToRecord);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToRecord = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstReceiver[nHighwayNo].nTsToRecord = nTsToRecord;

        g_hLogicDevs.mBurstReceiver[nHighwayNo].bActive = OTX_TRUE;
      } else {
        DemoLogf("Burst Receiver[%d] Thread already running.\n", nHighwayNo);
      }
    }
    break;

  case 'P':
    {
      OTX_UINT32 nHighwayNo, i;
      OTX_UINT32 nUserInput = 0;
      OTX_UINT32 *pUserInput;
      char strBuf[25];

      DemoPrintf("Enter fixed value (ie 0x12345678) : ");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nUserInput = strtol(strBuf, NULL, 0);
      }

      nHighwayNo = g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs;
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstSender[%d]. Continously send static values\n", nHighwayNo);

        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;

        // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf).
        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = 0;
        pUserInput = (OTX_UINT32 *) g_hLogicDevs.mBurstSender[nHighwayNo].sqTxBuf;
        for (i = 0; i < DEMO_BURST_BUF_SIZE/4; i++) {
            *pUserInput = nUserInput; // 0 - 31
            pUserInput++;
        }
        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;

        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = DEMO_BURST_BUF_SIZE;
      } else {
        DemoLogf("Burst Sender[%d] already running.\n", nHighwayNo);
      }
    }

    break;

    case 's':
    {
      OTX_UINT32 nHighwayNo, i, base;

      nHighwayNo = g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs;
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstSender[%d]. Continously send static values\n", nHighwayNo);

        g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = NULL;

        // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf).
        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = 0;

        if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_2_EXPRESSCARD) {
          // The Thor-2-ExpressCard has 128 timeslots in the burst highway. 
          // 32 of them are for Li0.
          // Another 32 of them are for Li1, and then we also have 64 
          // timeslots from the DSP device.
          // TS0, TS4, TS8, ...  ,TS124 are mapped to Li0
          // TS2, TS6, TS10, ... ,TS126 are mapped to Li1
          // TS1, TS3, TS5, ...  ,TS127 are mapped to the DSP
          int  x;
          char nByteData;
          char *pByte;

          DemoLogf("BurstSender[%d]. ExpressCard\n", nHighwayNo);

          nByteData = 0;
          for (x = 0; x < DEMO_BURST_BUF_SIZE; x = x + 4) {

            pByte = (OTX_CHAR *) &g_hLogicDevs.mBurstSender[nHighwayNo].sqTxBuf[x];
            *pByte = nByteData;        // Li0
            pByte++;
            *pByte = 0; // DSP not used
            pByte++;
            *pByte = nByteData | 0x10; // Li1
            pByte++;
            *pByte = 0; // DSP not used

            nByteData++;
            nByteData = (char) (nByteData % 10); // 0,1,2,3...9
          }
        }
        else
        if(g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_EXPRESSCARD) {
          // The Thor-4-ExpressCard has 128 timeslots in the burst highway. 
          // 32 of them are for Li0.
          // Another 32 of them are for Li1, and the 64 remaining timeslots 
          // are not used when sending data to the FALC.
          // TS0, TS4, TS8, ...  ,TS124 are mapped to Li0
          // TS1, TS5, TS9, ...  ,TS125 are mapped to Li1
          // TS2, TS6, TS10, ... ,TS126 Not used Li2 is receive only
          // TS3, TS7, TS17, ... ,TS127 Not used Li3 is receive only
          int  x;
          char nByteData;
          char *pByte;

          DemoLogf("BurstSender[%d]. ExpressCard\n", nHighwayNo);

          nByteData = 0;
          x = 0;
          pByte = (OTX_CHAR *) &g_hLogicDevs.mBurstSender[nHighwayNo].sqTxBuf[x];
          for ( ; x < DEMO_BURST_BUF_SIZE; x = x + 4) {
            *pByte = nByteData;        // Li0
            pByte++;
#if 0
            *pByte = nByteData | 0x80; // Li1
#else
            *pByte = nByteData | 0x10; // Li1
#endif
            pByte++;
            *pByte = 0; // Not used
            pByte++;
            *pByte = 0; // Not used
            pByte++;

            nByteData++;
#if 0
            nByteData = (char) (nByteData % 32); // 0,1,2,3...31
#else
            nByteData = (char) (nByteData % 10); // 0,1,2,3...9
#endif
          }
        }
        else {
#if 0 // For use with multiple boards.    
          base = ((g_demoEnv.coLiPerBurstDevs * nHighwayNo) * 0x10) + 
                 ((g_demoEnv.boardNo << 2) * 0x10);
#else
          base = ((g_demoEnv.coLiPerBurstDevs * nHighwayNo) * 0x10);
#endif
          for (i = 0; i < DEMO_BURST_BUF_SIZE; i++) {
            // buffer is laid out as (allows easy verification of data):
            //   LI0 [0]=0x00 [1]=0x01 .. [10]=0x00 [11]=0x01 .. [30]=0x00 [31]=0x01
            //   LI1 [0]=0x10 [1]=0x11 .. [10]=0x10 [11]=0x11 .. [30]=0x10 [31]=0x11
            //      ...
            //   LI6 [0]=0x60 [1]=0x61 .. [10]=0x60 [11]=0x61 .. [30]=0x60 [31]=0x61
            //   LI7 [0]=0x70 [1]=0x71 .. [10]=0x70 [11]=0x71 .. [30]=0x70 [31]=0x71
#if 0 
            g_hLogicDevs.mBurstSender[nHighwayNo].sqTxBuf[i] = (char) (i % 32); // 0 - 31
#else
            g_hLogicDevs.mBurstSender[nHighwayNo].sqTxBuf[i] = (char) (base + (((i % g_demoEnv.coTimeslotsOnBurstHwy) / 32) * 0x10) + ((i % 32) % 10));
#endif
          }
        }

        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;

        g_hLogicDevs.mBurstSender[nHighwayNo].cbTxBuf = DEMO_BURST_BUF_SIZE;
      } else {
        DemoLogf("Burst Sender[%d] already running.\n", nHighwayNo);
      }
    }

    break;

  case 'S':
    {
      char sqzBuf[OTX_MAX_PATH];
      OTX_UINT32 nHighwayNo;
      static OTX_INT32 nTsToSend = -1;

      nHighwayNo = g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs;
      if (g_hLogicDevs.mBurstSender[nHighwayNo].bActive == OTX_FALSE) {
        DemoLogf("BurstSender[%d]. Please enter file to send: ", nHighwayNo);

        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0] == '\0') {
          DemoLogf("File name not entered. Operation aborted\n");
          break;
        }

        if ((g_hLogicDevs.mBurstSender[nHighwayNo].pTxFile = fopen(sqzBuf, "rb")) == NULL) {
          DemoLogf("Error: Could not open file %s \n", sqzBuf);
          break;
        }

        DemoLogf("BurstSender[%d]. Please enter single TS[0-%d] to send from file (-1=All) [%d]: ", nHighwayNo, g_demoEnv.coTimeslotsOnBurstHwy - 1, nTsToSend);
        OsaConsoleGetStr(sqzBuf);
        if (sqzBuf[0]) {
          nTsToSend = strtol(sqzBuf, NULL, 0);
        }
        g_hLogicDevs.mBurstSender[nHighwayNo].nTsToSend = nTsToSend;

        g_hLogicDevs.mBurstSender[nHighwayNo].bActive = OTX_TRUE;
        g_hLogicDevs.mBurstSender[nHighwayNo].bVerbose = OTX_TRUE;
      } else {
        DemoLogf("Burst Sender[%d] already running.\n", nHighwayNo);
      }
    }
    break;

  case 'L':
    g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bFileLoop = !g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bFileLoop;
    DemoLogf("Setting loop file sends %s for Burst Sender[%d] Thread\n", (g_hLogicDevs.mBurstSender[g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs].bFileLoop)?"ON":"OFF", g_demoEnv.nCurrentLi / g_demoEnv.coLiPerBurstDevs);
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;

  case 'p':
    {
      OtxT1E1LoopE eLoopType;
      OTX_INT32 nLi;

      DemoLogf("Set Loopback mode for all Line Interfaces \n");

      DemoLogf("0: All Line Loops off \n");
      DemoLogf("1: Line Loop (External highway loop) \n");
      DemoLogf("2: Remote Loop with Jitter Attenuation On (External highway loop)\n");
      DemoLogf("3: Remote Loop with Jitter Attenuation Off (External highway loop)\n");
      DemoLogf("4: Local Loop (Internal highway loop) \n");
      
      switch (OsaConsoleGetCh() - '0') {
      case 0:
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      case 1:
        eLoopType = OTX_T1E1_LINE_LOOP;
        break;
      case 2:
        eLoopType = OTX_T1E1_REMOTE_ON_LOOP_JITTER_ATTEN;
        break;
      case 3:
        eLoopType = OTX_T1E1_REMOTE_OFF_LOOP_JITTER_ATTEN;
        break;
      case 4:
        eLoopType = OTX_T1E1_LOCAL_LOOP;
        break;
      default:
        DemoLogf("Invalid entry. Turning Loop off (OTX_T1E1_LOOPS_OFF)\n");
        eLoopType = OTX_T1E1_LOOPS_OFF;
        break;
      }

      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        DemoLogf("Setting Loopback mode (%d) on Li #%d \n", (int)eLoopType, nLi);
        result = OtxT1E1SetLineLoop(g_hPhysDevs.hLi[nLi], eLoopType); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
      }

      DemoLogf("Done.\n");
    }
    break;

  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    if (commandChar - '0' < g_demoEnv.coLi) {
      // Select current Li
      g_demoEnv.nCurrentLi = commandChar - '0';
      DemoLogf("LI#%d Selected ....\n", g_demoEnv.nCurrentLi);
    }
    break;

  case 'T':
    // reset stats
    OtxDrvDbgDump(g_hPhysDevs.hBurst, 101);
    break;

  case 'b':
    {
      char strBuf[16];
      OTX_INT32 nLoopCnt = 0;
      OTX_INT32 nCntWithCmd = 0;

      DemoPrintf("Enter number of loop delay count :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nLoopCnt = strtol(strBuf, NULL, 0);
      }
      // Only have 16 bits
      nLoopCnt = nLoopCnt & 0xFFFF;

      DemoPrintf("Loop count is %d \n", nLoopCnt);

      // Setup the command. Loop count is upper 16 bits
      nCntWithCmd = (nLoopCnt << 16) | 0x1000;
      
      // dump stats to debugger
      OtxDrvDbgDump(g_hPhysDevs.hThorController, nCntWithCmd);
    }
    break;

  case 'B':
    {
      char strBuf[16];
      OTX_INT32 nLoopCnt = 0;
      OTX_INT32 nCntWithCmd = 0;

      DemoPrintf("Enter number of loop delay count :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nLoopCnt = strtol(strBuf, NULL, 0);
      }
      // Only have 16 bits
      nLoopCnt = nLoopCnt & 0xFFFF;

      DemoPrintf("Loop count is %d \n", nLoopCnt);

      // Setup the command. Loop count is upper 16 bits
      nCntWithCmd = (nLoopCnt << 16) | 0x1001;
      
      // dump stats to debugger
      OtxDrvDbgDump(g_hPhysDevs.hThorController, nCntWithCmd);
    }
    break;

  case 'N':
    {
      OTX_UINT32 nValue;
      char sValue[80];

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      DemoLogf("Enter queue notification value (1-32) [currently %d]: ", nValue);
      OsaConsoleGetStr(sValue);
      if (sValue[0] == '\0') {
        break;
      }
      nValue = strtol(sValue, NULL, 0);

      // before setting the new trigger value, the burst device must not be enabled
      result = OtxDrvDisable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      result = OtxDrvEnable(g_hPhysDevs.hBurst,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hBurst, result);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_TRIGGER, &nValue); DemoCheckRc(g_hPhysDevs.hBurst, result);
      DemoLogf("New queue notification value is %d\n", nValue);

      break;
    }

  case 'C':
    {
      OTX_UINT8 highWay;
      OTX_UINT8 channel;
      OTX_UINT8 byteVal;
      char wValStr[12];
      OTX_RESULT result;
      
      if(g_hPhysDevs.hTss != OTX_INVALID_HANDLE_VALUE) {
        DemoLogf("Set Constant Byte on Channel\n");
      
        DemoLogf("Highway: ");
        OsaConsoleGetStr(wValStr);
        highWay = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
        DemoLogf("Channel: ");
        OsaConsoleGetStr(wValStr);
        channel = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
        DemoLogf("Constant Byte: ");
        OsaConsoleGetStr(wValStr);
        byteVal = (OTX_UINT8) strtoul(wValStr, NULL, 0);
      
        result = OtxBrdTssConstByte(g_hPhysDevs.hTss, highWay, channel, byteVal); 
        DemoCheckRc(g_hPhysDevs.hTss, result);
      }
      else {
        DemoLogf("Board does not support TSS device\n");
      }

      DemoLogf("Done.\n");
    }
    break;

  case 'z':
    {

      OTX_UINT32 stat_Starts           = g_nRcvStat_Starts;            // Number of actual dma transfers initiated
      OTX_UINT32 stat_Completes        = g_nRcvStat_Completes;         // Number of actual dma transfers completed IRQs
      OTX_UINT32 stat_Partial          = g_nRcvStat_Partial;           // Number of partial transfer due to framemask
      OTX_UINT32 stat_Full             = g_nRcvStat_Full;              // Number of count of max dma size transfers
      OTX_UINT32 stat_Aborts           = g_nRcvStat_Aborts;            // Number of actual dma transfers aborts
      OTX_UINT32 stat_Amount           = g_nRcvStat_Amount;            // Number of amount of data transfered
      OTX_UINT32 stat_Errors           = g_nRcvStat_Errors;            // Number of count of processing errors
      OTX_UINT32 stat_OUruns           = g_nRcvStat_OUruns;            // Number of amount of over/under runs
      OTX_UINT32 stat_OUrunsUser       = g_nRcvStat_OUrunsUser;        // Number of amount of over/under runs caused by user mode
      OTX_UINT32 stat_DmaBusy          = g_nRcvStat_DmaBusy;           // Number of count of DMA busy occurances
      OTX_UINT32 stat_Notify           = g_nRcvStat_Notify;            // Number of logical device notifies
      OTX_UINT32 stat_NotifyPartial    = g_nRcvStat_NotifyPartial;     // Number of logical device notifies - partial
      OTX_UINT32 stat_FifoSize         = g_nRcvStat_FifoSize;          // size of the Burst FIFO
      OTX_UINT32 stat_FifoTrigger      = g_nRcvStat_FifoTrigger;       // the current FIFO Trigger value
      OTX_UINT32 stat_FifoCount        = g_nRcvStat_FifoCount;         // the current amount in the FIFO
      OTX_UINT32 stat_FifoPeak         = g_nRcvStat_FifoPeak;          // the peak FIFO usage
      OTX_UINT32 stat_FifoTriggerCount = g_nRcvStat_FifoTriggerCount;  // the current FIFO trigger fire value
      
      if (!g_bReadBurstStatsThreadActive) {
        DemoLogf("Starting Stats Thread\n");
        g_bReadBurstStatsThreadActive = !g_bReadBurstStatsThreadActive;
        OsaThreadStart(DemoReadBurstStatsThread);
        OsaThreadSleep(1000);
      }

      DemoLogf("\n");

      DemoLogf("Burst Recv::\n Starts:\t0x%X,\n Completes:\t0x%X,\n Partial:\t0x%X,\n Full\t\t%d,\n Aborts:\t%d,\n Amount:\t0x%X,\n",
          stat_Starts,
          stat_Completes,
          stat_Partial,
          stat_Full,
          stat_Aborts,
          stat_Amount
      );

      DemoLogf(" Errors:\t%d,\n OUErrors:\t%d,\n OUErrUser:\t%d,\n DMABusy:\t%d,\n Notify:\t%d,\n",
          stat_Errors,
          stat_OUruns,
          stat_OUrunsUser,
          stat_DmaBusy,
          stat_Notify
      );

      DemoLogf(" NotifyPartial:\t%d,\n FIFO Size:\t%d,\n Trigger:\t%d,\n Count:\t\t%d,\n Peak:\t\t%d,\n TriggerCount:\t%d\n",
          stat_NotifyPartial,
          stat_FifoSize,
          stat_FifoTrigger,
          stat_FifoCount,
          stat_FifoPeak,
          stat_FifoTriggerCount
      );

      DemoLogf("\n");
    }
    break;

  case 'Z':
    {
      if (!g_bReadBurstStatsThreadActive) {
        DemoLogf("Starting Stats Thread\n");
        g_bReadBurstStatsThreadActive = !g_bReadBurstStatsThreadActive;
        OsaThreadStart(DemoReadBurstStatsThread);
        OsaThreadSleep(1000);
      }

      DemoLogf("\n");
      DemoLogf("Burst Send::\n Starts:\t0x%X,\n Completes:\t0x%X,\n Partial:\t0x%X,\n Full\t\t%d,\n Aborts:\t%d,\n Amount:\t0x%X,\n",
          g_nSendStat_Starts,
          g_nSendStat_Completes,
          g_nSendStat_Partial,
          g_nSendStat_Full,
          g_nSendStat_Aborts,
          g_nSendStat_Amount
      );
      DemoLogf(" Errors:\t%d,\n OUErrors:\t%d,\n OUErrUser:\t%d,\n DMABusy:\t%d,\n Notify:\t%d,\n",
          g_nSendStat_Errors,
          g_nSendStat_OUruns,
          g_nSendStat_OUrunsUser,
          g_nSendStat_DmaBusy,
          g_nSendStat_Notify
      );
 
      DemoLogf(" NotifyPartial:\t%d,\n FIFO Size:\t%d,\n Trigger:\t%d,\n Count:\t\t%d,\n Peak:\t\t%d,\n TriggerCount:\t%d\n",
          g_nSendStat_NotifyPartial,
          g_nSendStat_FifoSize,
          g_nSendStat_FifoTrigger,
          g_nSendStat_FifoCount,
          g_nSendStat_FifoPeak,
          g_nSendStat_FifoTriggerCount
      );

      DemoLogf("\n");

    }
    break;

  // Debug register reads

  case 'i':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      int nCurrentLi = g_demoEnv.nCurrentLi;

      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in QuadE1T1%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hQuadLi[nCurrentLi], 
                              nAddress, &readVal); 
      DemoCheckRc(g_hPhysDevs.hQuadLi[nCurrentLi], result);

      DemoLogf("hQuadLi%d Address 0x%X = %.2X\n", 
                nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;
  case 'I':
    {
      char strBuf[12];
      int nCurrentLi = g_demoEnv.nCurrentLi;

      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in QuadE1T1%d \n", 
               nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hQuadLi[nCurrentLi], 
                               nAddress, nValue); 
      DemoCheckRc(g_hPhysDevs.hQuadLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to hQuadLi%d Address 0x%X\n", 
               (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'j':
    {
      OTX_UINT32 readVal;
      char strBuf[12];

      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevs.hThorController, nAddress, &readVal); 
      DemoCheckRc(g_hPhysDevs.hThorController, result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;
  case 'J':
    {
      char strBuf[12];

      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevs.hThorController, nAddress, nValue); 
      DemoCheckRc(g_hPhysDevs.hThorController, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;
  case 'k':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      int nCurrentLi = g_demoEnv.nCurrentLi;
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); 
      DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;
  case 'K':
    {
      char strBuf[12];
      int nCurrentLi = g_demoEnv.nCurrentLi;

      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); 
      DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;

  case 'm':
    {
      OTX_UINT32 readVal;
      char strBuf[12];

      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 32-bit register in the board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg32(g_hPhysDevs.hThorController, nAddress, &readVal); 
      DemoCheckRc(g_hPhysDevs.hThorController, result);

      DemoLogf("Brd Address 0x%X = %.8X\n", nAddress, readVal);
    }
    break;
  case 'M':
    {
      char strBuf[12];

      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 32-bit register in the board device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg32(g_hPhysDevs.hThorController, nAddress, nValue); 
      DemoCheckRc(g_hPhysDevs.hThorController, result);

      DemoLogf("Wrote %.8X to Brd Address 0x%X\n", nValue, nAddress);
    }
    break;

  case '+':
    // DUMP falc56 registers
    OtxDrvDbgDump(g_hPhysDevs.hThorController, 100);
    break;

  case '-':
    // Set transparent mode
    OtxDrvDbgDump(g_hPhysDevs.hThorController, 101);
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      DemoPrintf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        DemoPrintf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coLi              = 8;
  envP->coHighways        = 2;
  envP->coLiPerBurstDevs  = 4;
  envP->coTimeslotsOnBurstHwy = 128;
  envP->nCurrentLi        = 0;
  envP->bDisableBurstTx   = OTX_FALSE;
  envP->bDisableBurstRx   = OTX_FALSE;

  envP->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eTermMode         = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '4') {
            envP->nBoardType = OTX_DEVICE_THOR_4_PCI_PLUS;
            envP->coLi = 4;
            envP->coHighways = 1;
            envP->coLiPerBurstDevs = 4;
            envP->coTimeslotsOnBurstHwy = 128;
          } else if ( *(argv[i]+2) == 'P') {
            envP->nBoardType = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 8;
            envP->coHighways = 2;
            envP->coLiPerBurstDevs = 4;
            envP->coTimeslotsOnBurstHwy = 128;
          } else if ( *(argv[i]+2) == 'T') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_PLUS;
            envP->coLi = 2;
            envP->coHighways = 2;
            envP->coLiPerBurstDevs = 1;
            envP->coTimeslotsOnBurstHwy = 32;
          } else if ( *(argv[i]+2) == 'E') {
            envP->nBoardType = OTX_DEVICE_THOR_2_PCI_EXPRESS;
            envP->coLi = 2;
            envP->coHighways = 2;
            envP->coLiPerBurstDevs = 1;
            envP->coTimeslotsOnBurstHwy = 32;
          } else if ( *(argv[i]+2) == 'S') {
            envP->nBoardType = OTX_DEVICE_SLEIPNIR_1_PCI;
            envP->coLi = 1;
            envP->coHighways = 1;
            envP->coLiPerBurstDevs = 1;
            envP->coTimeslotsOnBurstHwy = 32;
          } else if ( *(argv[i]+2) == 'H') {
            envP->nBoardType = OTX_DEVICE_HUGIN_2_PC104;
            envP->coLi = 2;
            envP->coHighways = 2;
            envP->coLiPerBurstDevs = 1;
            envP->coTimeslotsOnBurstHwy = 32;
          } else if ( *(argv[i]+2) == '1') {
            envP->nBoardType = OTX_DEVICE_THOR_1_PCI_PLUS;
            envP->coLi = 1;
            envP->coHighways = 1;
            envP->coLiPerBurstDevs = 1;
            envP->coTimeslotsOnBurstHwy = 32;
          } else if ( *(argv[i]+2) == 'C') {
            envP->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
            // The Thor-2-ExpressCard has 128 timeslots in the burst highway. 32 of them are for Li0.
            // Another 32 of them are for Li1, and then we also have 64 timeslots from the DSP device.
            // TS0, TS4, TS8, ...  ,TS124 are mapped to Li0
            // TS2, TS6, TS10, ... ,TS126 are mapped to Li1
            // TS1, TS3, TS5, ...  ,TS127 are mapped to the DSP
            envP->coLi = 2;
            envP->coHighways = 1;
            envP->coLiPerBurstDevs = 2;
            envP->coTimeslotsOnBurstHwy = 128;
          } else if ( *(argv[i]+2) == 'R') {
            envP->nBoardType = OTX_DEVICE_THOR_4_EXPRESSCARD;
            // The Thor-4-ExpressCard has 128 timeslots in the burst highway. 32 
            // of them are for Li0, Li1, Li2 and Li3. Li2 and Li3 are receive 
            // only used to monitor 2 spans.
            // TS0, TS4, TS8,  ... ,TS124 are mapped to Li0
            // TS1, TS5, TS9,  ... ,TS125 are mapped to Li1
            // TS2, TS6, TS10, ... ,TS126 are mapped to Li2 (receive only)
            // TS3, TS7, TS11, ... ,TS127 are mapped to Li3 (receive only)
            envP->coLi = 4;
            envP->coHighways = 1;
            envP->coLiPerBurstDevs = 2;
            envP->coTimeslotsOnBurstHwy = 128;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'D') {   // Disable one direction of burst
          if (*(argv[i]+2) == 'T') {   // TX
            envP->bDisableBurstTx = OTX_TRUE;
          } else if (*(argv[i]+2) == 'R') {   // T1
            envP->bDisableBurstRx = OTX_TRUE;
          } else {
            DemoLogf("Error: Invalid Burst direction-disabling command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if ( (*(argv[i]+2) == 'H') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS_ETCF;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'N') {
            envP->eT1FrameFormat = OTX_T1_FF_UNFRAMED; // Transparent mode
          } else if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'N') {
            envP->eE1FrameFormat = OTX_E1_FF_UNFRAMED; // Transparent mode
          } else if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[T|E|4|P|S|H|I|C|R]    : Board Type used for the demo, where\n");
  DemoLogf("                  T = Use Thor-2-PCI-Plus Board\n");
  DemoLogf("                  E = Use Thor-2-PCI-Express Board\n");
  DemoLogf("                  4 = Use Thor-4-PCI-Plus Board\n");
  DemoLogf("                  P = Use Thor-8-PCI-Plus Board\n");
  DemoLogf("                  S = Use Sleipnir Board\n");
  DemoLogf("                  H = Use Hugin-2-PC104 Board\n");
  DemoLogf("                  1 = Use Thor-1-PCI-Plus Board\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board\n");
  DemoLogf("                  R = Use Thor-4-ExpressCard Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /D[T|R]      : Disable one direction of the Burst device\n");
  DemoLogf("                  T = Disable Burst TX\n");
  DemoLogf("                  R = Disable Burst RX\n");
  DemoLogf(" /O[x]        : Clock Source, where\n");
  DemoLogf("                  x = I Internal clock source (default) or\n");
  DemoLogf("                  x = 0 through 7 indicating with Li to recover the clock from or\n");
  DemoLogf("                  x = H to recover from H.100 A-clock (ETCF mode)\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced (default)\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /F[N|E|S|9]  : T1 Frame Format, where\n");
  DemoLogf("                  N = Not framed/Unframed\n");
  DemoLogf("                  E = ESF (default)\n");
  DemoLogf("                  4 = F4\n");
  DemoLogf("                  S = SF [D3/D4] (F12)\n");
  DemoLogf("                  9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]      : T1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  B = B8ZS (default)\n");
  DemoLogf(" /G[N|D|C|G]  : E1 Frame Format, where\n");
  DemoLogf("                  N = Not framed/Unframed\n");
  DemoLogf("                  D = Doubleframe\n");
  DemoLogf("                  C = CRC4 Multiframe (default)\n");
  DemoLogf("                  G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]      : E1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  H = HDB3 (default)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile,   formatStr, argList);

    fflush(g_pDemoLogFile);
  }
  va_end(argList);

}


/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  OTX_INT32 i;

  for (i = 0; i < g_demoEnv.coHighways; i++) {
    if (g_hLogicDevs.mBurstSender[i].pTxFile) {
      fclose(g_hLogicDevs.mBurstSender[i].pTxFile);
      g_hLogicDevs.mBurstSender[i].pTxFile = NULL;
    }
    if (g_hLogicDevs.mBurstReceiver[i].pRxFile) {
      fclose(g_hLogicDevs.mBurstReceiver[i].pRxFile);
      g_hLogicDevs.mBurstReceiver[i].pRxFile = NULL;
    }
  }

  if (g_pDemoLogFile) {
    fclose(g_pDemoLogFile);
  }
}

int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  OsaConsoleInit();

  printf("T8pBurstEx, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program for Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);


  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
//    OsaConsoleExit();
//    exit(-1);
  }
  atexit(DemoAtExit);

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board
  {
    char szClockSource[64];

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    case OTX_CLOCK_SOURCE_A_CLOCKS_ETCF:
      sprintf(szClockSource, "H.100 A-Clocks (ETCF mode)");
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);

    result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_H100_ETCF, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController, result);
  }

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController, result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\n");
  DemoLogf("Press 'h' for help\n");

  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // abort threads
  DemoLogf("\n");
  DemoLogf("Exiting...\n");
  {
    OTX_INT32 i;

    g_bStatsThreadActive = OTX_FALSE;
    g_bReadBurstStatsThreadActive = OTX_FALSE;
    for (i = 0; i < g_demoEnv.coHighways; i++) {
      g_hLogicDevs.mBurstReceiver[i].bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender[i].bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}
